# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""Functions to help enable math and fitting in the main examine code"""

from __future__ import print_function, division

import numpy as np
import warnings
from astropy.modeling import models, fitting




def gfwhm(sigmax, sigmay=None):
    """Compute the gaussian full width half max.

    Parameters
    ----------

    sigmax: float
        The input sigma to use

    sigmay: float
        The input sigma to use

    Returns
    -------
    The FWHM tuple for the gaussian

    """
    if sigmax is None:
        print("Need at least one sigma value")
        return (None, None)

    g = lambda x: x * np.sqrt(8.0 * np.log(2.))

    if sigmay is None:
        return (g(sigmax), g(sigmax))  # assume circular where sigmax = sigmay
    else:
        return (g(sigmax), g(sigmay))


def mfwhm(alpha=0, gamma=0):
    """Compute the Moffat full width half max.

    Parameters
    ----------

    sigmax: float
        The input sigma to use

    sigmay: float
        The input sigma to use

    Returns
    -------
    The FWHM tuple for a Moffat function
    fwhm = 2* alpha * sqrt(2^(1/gamma) -1 ))
    """
    if alpha == 0 or gamma == 0:
        print("Need alpha AND gamma values")
        return None

    g = lambda x, y: 2 * x * np.sqrt(2 ** (1/y) - 1)

    return g(alpha, gamma)


def fit_moffat_1d(data, gamma=2., alpha=1.):
    """Fit a 1D moffat profile to the data and return the fit."""
    # data is assumed to already be chunked to a reasonable size
    ldata = len(data)
    x = np.arange(ldata)

    # Fit model to data
    fit = fitting.LevMarLSQFitter()

    # Moffat1D
    model = models.Moffat1D(amplitude=max(data), x_0=ldata/2, gamma=gamma, alpha=alpha)
    with warnings.catch_warnings():
        # Ignore model linearity warning from the fitter
        warnings.simplefilter('ignore')
        results = fit(model, x, data)

    # previous yield amp, ycenter, xcenter, sigma, offset
    return results


def fit_gauss_1d(data):
    """Fit a 1D gaussian to the data and return the fit."""
    # data is assumed to already be chunked to a reasonable size
    ldata = len(data)
    x = np.arange(ldata)

    # Fit model to data
    fit = fitting.LevMarLSQFitter()

    # Gaussian1D
    model = models.Gaussian1D(amplitude=1, mean=0, stddev=1.)
    with warnings.catch_warnings():
        # Ignore model linearity warning from the fitter
        warnings.simplefilter('ignore')
        results = fit(model, x, data)

    # previous yield amp, ycenter, xcenter, sigma, offset
    return results


def gauss_center(data, sigma=3., theta=0.):
    """center the data  by fitting a 2d gaussian to the region.

    Parameters
    ----------

    data: float
        should be a 2d array, the initial center is used to estimate
        the fit center
    """
    # use a smaller bounding box so that we are only fitting the local data
    delta = int(len(data) / 2)  # guess the center
    ldata = len(data)
    yy, xx = np.mgrid[:ldata, :ldata]

    # Fit model to data
    fit = fitting.LevMarLSQFitter()

    # Gaussian2D(amp,xmean,ymean,xstd,ystd,theta)
    model = models.Gaussian2D(1, delta, delta, sigma, sigma, theta)
    with warnings.catch_warnings():
        # Ignore model linearity warning from the fitter
        warnings.simplefilter('ignore')
        results = fit(model, xx, yy, data)

    # previous yield amp, ycenter, xcenter, sigma, offset
    return results


def fit_mex_hat_1d(data):
    """Fit a 1D Mexican Hat function to the data."""
    ldata = len(data)
    x = np.arange(ldata)
    fixed_pars = {"x_0": True}

    # Fit model to data
    fit = fitting.LevMarLSQFitter()

    # Mexican Hat 1D
    model = models.MexicanHat1D(amplitude=np.max(data),
                                x_0=ldata/2, sigma=2., fixed=fixed_pars)
    with warnings.catch_warnings():
        # Ignore model linearity warning from the fitter
        warnings.simplefilter('ignore')
        results = fit(model, x, data)

    # previous yield amp, ycenter, xcenter, sigma, offset
    return results


def fit_airy_2d(data, x=None, y=None):
    """Fit an AiryDisk2D model to the data."""
    delta = int(len(data) / 2)  # guess the center
    ldata = len(data)

    if not x:
        x = delta
    if not y:
        y = delta
    fixed_pars = {"x_0": True, "y_0": True}  # hold these constant
    yy, xx = np.mgrid[:ldata, :ldata]

    # fit model to the data
    fit = fitting.LevMarLSQFitter()

    # AiryDisk2D(amplitude, x_0, y_0, radius)
    model = models.AiryDisk2D(np.max(data), x_0=x, y_0=y, radius=delta,
                              fixed=fixed_pars)
    with warnings.catch_warnings():
            # Ignore model warnings for new_plot_window
            warnings.simplefilter('ignore')
            results = fit(model, xx, yy, data)

    return results


def fit_poly_n(data, x=None, y=None, deg=1):
    """Fit a Polynomial 1D model to the data."""

    # define the model
    poly = models.Polynomial1D(deg)

    # set the axis range for fitting
    ax = np.arange(len(data))

    # define the fitter
    fit = fitting.LinearLSQFitter()
    try:
        result = fit(poly, ax, data)
    except:
        ValueError
        result = None
    return result



"""Licensed under a 3-clause BSD style license - see LICENSE.rst."""

from __future__ import print_function, division

import os
import sys
import logging
import warnings
from astropy.io import fits

try:
    import xpa
    _have_xpa = True
except ImportError:
    _have_xpa = False

from .version import version as __version__

# To guide any import *
__all__ = ["display_help", "set_logging"]


def find_ds9():
    """Find the local path to the DS9 executable"""
    path = "ds9"
    for dirname in os.getenv("PATH").split(":"):
        possible = os.path.join(dirname, path)
        if os.path.isfile(possible):
            return possible
    return None


def find_xpans():
    """Find the local path to the xpans executable"""
    path = "xpans"
    for dirname in os.getenv("PATH").split(":"):
        possible = os.path.join(dirname, path)
        if os.path.exists(possible):
            return possible
    return None


def list_active_ds9(verbose=True):
    """
    Display and/or return information about the DS9 windows currently
    registered with the XPA.

    Parameters
    ----------
    verbose : bool
        If True, prints out all the information about what DS9 windows
        are active.

    Returns
    -------
    session_list : list
        The list of sessions that have been registered.  Each entry in the list
        is a list containing the information that xpans yields.  Typically the
        fourth element in that tuple contains the actual target name.

    Notes
    -----
    when I start a unix socket with connect() the xpa register isn't
    seeing it when I call this function. I think because it's only
    listening on the inet socket which starts by default in the OS.
    That's if xpans is installed on the machine. Otherwise, no
    nameserver is running at all. This helps with object cont
    """
    session_dict = {}

    # only run if XPA/xpans is installed on the machine
    if find_xpans():
        sessions = None
        try:
            sessions = xpa.get(b"xpans").decode().strip().split("\n")
            if ((sessions is None or len(sessions) < 1) and verbose):
                print("No active sessions")
            else:
                for line in sessions:
                    classn, name, access, ids, user = tuple(line.split())
                    session_dict[ids] = (name, user, classn, access)
                if verbose:
                    for line in sessions:
                        print(line)
        except xpa.XpaException:
                print("No active sessions registered")

    else:
        print("XPA nameserver not installed or not on PATH, \
               function unavailable")

    return session_dict


def display_help():
    """Display RTD html help for the installed verison in a browser window."""
    url = "http://imexam.readthedocs.io/"
    try:
        import webbrowser
        # grab the version that's installed
        if "dev" not in __version__:
            url += "en/v{0:s}/".format(__version__)
        webbrowser.open(url)
    except ImportError:
        warnings.warn(
            "webbrowser module not installed, see {0:s} for help".format(url))
        raise ImportError


def display_xpa_help():
    """Display help for XPA Acccess Points."""
    url = "http://ds9.si.edu/doc/ref/xpa.html"
    try:
        import webbrowser
        webbrowser.open(url)
    except ImportError:
        warnings.warn(
            "webbrowser module not installed, see {0:s} for help".format(url))
        raise ImportError


# Set up logging ability for the user
# consider making a private logging level for data retension
def set_logging(filename=None, on=True, level=logging.INFO):
    """Turn on or off logging to file or stdout.

    Parameters
    ----------

    filename: str, optional
        name of the file for logging information
    on: bool, optional
        turn logging on or off, will close the file output
        in order to turn off stdout logging, set the level
        to logging.CRITICAL
    level: logging, optional
        the logging level to be recorded or displayed

    """

    formatter = logging.Formatter('\n%(funcName)s \n%(message)s')
    root = logging.getLogger(__name__)
    root.setLevel(level)
    stream_attached = False
    file_attached = False

    if on:
        #  Try to avoid adding duplicate file handlers
        if len(root.handlers) > 0:
            for handler in root.handlers:
                if isinstance(handler, logging.StreamHandler):
                    stream_attached = True
                    handler.setLevel(level)
                if isinstance(handler, logging.FileHandler):
                    file_attached = True
                    raise ValueError("File for logging already specified,\
                                      turn off logging first.")
        else:
            # to prevent warning in unhandled contexts and messages to stderr
            root.addHandler(logging.NullHandler())

        if isinstance(filename, str) and not file_attached:
            file_handler = logging.FileHandler(filename=filename,
                                               mode='a',
                                               delay=True)
            file_handler.setLevel(logging.INFO)
            file_handler.setFormatter(formatter)
            print("Saving imexam commands to {0:s}".format(repr(filename)))
            root.addHandler(file_handler)

        if not stream_attached:
            # set the stdout stream handler
            stdout_handler = logging.StreamHandler(stream=sys.stdout)
            stdout_handler.setLevel(logging.INFO)
            # stdout_handler.setFormatter(formatter)
            root.addHandler(stdout_handler)

    #  turning the logging off to the file and set level on stream handler
    else:
        for handler in root.handlers:
            # close the file logger
            if isinstance(handler, logging.FileHandler):
                handler.close()
                root.removeHandler(handler)
            # set stream logging level
            if isinstance(handler, logging.StreamHandler):
                handler.setLevel(level)

    return root


def check_filetype(filename=None):
    """Check the file to see if it is a multi-extension FITS file
    or a simple fits image where the data and header are stored in
    the global unit.

    Parameters
    ----------
    filename: string
        The name of the file to check


    """
    log = logging.getLogger(__name__)

    if filename is None:
        raise ValueError("No filename provided")
    else:
        try:
            mef_file = fits.getval(filename, ext=0, keyword='EXTEND')
        except KeyError:
            mef_file = False
        except IOError:
            log.warning("Problem opening file {0:s}".format(repr(filename)))
            raise IOError("Error opening {0:s}".format(filename))

        #  double check for lying liars, should at least have 1 extension
        #  and XTENSION is a required keyword
        if mef_file:
            try:
                fits.getval(filename, ext=1, keyword='XTENSION')
            except (KeyError, IndexError):
                mef_file = False
        return mef_file


def verify_filename(filename=None, extver=None, extname=None):
    """Verify the filename exist and split it for extension information.
    If the user has given an extension, extension name or some combination of
    those, return the full filename, extension and version tuple.

    Parameters
    ----------
    filename: string
        The name of the file to verify

    extver: int
        extsion number which corresponds to extname

    extname: string
        the name of the extension
    """
    if filename is None:
        print("No filename provided")
    else:
        try:
            if "[" in filename:
                splitstr = filename.split("[")
                shortname = splitstr[0]
                if "," in splitstr[1]:
                    extname = splitstr[1].split(",")[0]
                    extver = int(splitstr[1].split(",")[1][0])
                else:
                    extver = int(filename.split("[")[1][0])
            else:
                shortname = os.path.abspath(filename)
        except IndexError as e:
            print("Exception: {0}".format(repr(e)))
            raise IndexError

        return shortname, extname, extver
